﻿using System;
using Dynamicweb.Ecommerce.Orders;

namespace Dynamicweb.Ecommerce.LiveIntegration.Extensions
{
  /// <summary>
  /// Extension methods for order lines.
  /// </summary>
  internal static class OrderLineExtensions
  {
    /// <summary>
    /// Determines if this order line represents a product (and not a discount or tax line for example).
    /// </summary>
    /// <param name="orderLine">The order line.</param>
    /// <returns>Returns <c>true</c> if the specified order line is a product; otherwise, <c>false</c>.</returns>
    public static bool IsProduct(this OrderLine orderLine)
    {
      return orderLine.Type == "0" || orderLine.Type == "2";
    }

    /// <summary>
    /// Determines if this order line represents a discount (and not a product or tax line for example).
    /// </summary>
    /// <param name="orderLine">The order line.</param>
    /// <returns>Returns <c>true</c> if the specified order line is a discount; otherwise, <c>false</c>.</returns>
    public static bool IsDiscount(this OrderLine orderLine)
    {
      return orderLine.Type == "1" || orderLine.Type == "3";
    }

    /// <summary>
    /// Determines if this order line represents a product discount (and not a product or tax line for example).
    /// </summary>
    /// <param name="orderLine">The order line.</param>
    /// <returns>Returns <c>true</c> if the specified order line is a product discount; otherwise, <c>false</c>.</returns>
    public static bool IsProductDiscount(this OrderLine orderLine)
    {
      return orderLine.Type == "3";
    }

    /// <summary>
    /// Determines if this order line represents an order discount (and not a product or tax line for example).
    /// </summary>
    /// <param name="orderLine">The order line.</param>
    /// <returns>Returns <c>true</c> if the specified order line is an order discount; otherwise, <c>false</c>.</returns>
    public static bool IsOrderDiscount(this OrderLine orderLine)
    {
      return orderLine.Type == "1";
    }

    /// <summary>
    /// Determines if this order line represents a tax line (and not a product or discount line for example).
    /// </summary>
    /// <param name="orderLine">The order line.</param>
    /// <returns>Returns <c>true</c> if the specified order line is a tax line; otherwise, <c>false</c>.</returns>
    public static bool IsTax(this OrderLine orderLine)
    {
      return orderLine.Type == "4";
    }

    /// <summary>
    /// Gets the order line type description.
    /// </summary>
    /// <param name="orderLine">The order line.</param>
    /// <returns>System.String.</returns>
    /// <exception cref="NotSupportedException"></exception>
    public static string GetOrderLineTypeDescription(this OrderLine orderLine)
    {
      if (orderLine.IsProduct())
      {
        return "Product";
      }

      if (orderLine.IsDiscount())
      {
        return "Discount";
      }

      if (orderLine.IsTax())
      {
        return "Tax";
      }

      throw new NotSupportedException($"Unsupported type {orderLine.Type}");
    }

    /// <summary>
    /// Clones the order line.
    /// </summary>
    /// <param name="orderLine">The order line.</param>
    /// <returns>OrderLine.</returns>
    internal static OrderLine CloneOrderLine(this OrderLine orderLine)
    {
      OrderLine newOrderLine = new OrderLine
      {
        Order = orderLine.Order,
        Quantity = orderLine.Quantity,
        Type = orderLine.Type,
        ProductName = orderLine.ProductName,
        ParentLineId = orderLine.ParentLineId,
        UnitPrice =
                {
                    PriceWithVAT = orderLine.UnitPrice.PriceWithVAT,
                    PriceWithoutVAT = orderLine.UnitPrice.PriceWithoutVAT,
                    VAT = orderLine.UnitPrice.VAT
                }
      };

      newOrderLine.Price.PriceWithVAT = orderLine.Price.PriceWithVAT;
      newOrderLine.Price.PriceWithoutVAT = orderLine.Price.PriceWithoutVAT;
      newOrderLine.DiscountId = orderLine.DiscountId;

      return newOrderLine;
    }
  }
}