﻿using Dynamicweb.Ecommerce.LiveIntegration.Extensions;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Orders.SalesDiscounts;
using Dynamicweb.Extensibility.AddIns;
using System;
using System.Linq;

namespace Dynamicweb.Ecommerce.LiveIntegration
{
    /// <summary>
    /// Sales Discount Provider to handle discount lines coming from the ERP.
    /// </summary>
    /// <seealso cref="SalesDiscountProvider" />
    [AddInName("Live integration discount"),
    AddInDescription("Apply discounts retrieved via live integration")]
    public class LiveSalesDiscountProvider : SalesDiscountProvider
    {
        /// <summary>
        /// Gets or sets a value indicating whether to hide discount value group box in the UI.
        /// </summary>
        /// <value><c>true</c> if it is needed to hide discount value group box; otherwise, <c>false</c>.</value>
        public override bool HideDiscountValueGroupBox
        {
            get
            {
                return true;
            }

            set
            {
                base.HideDiscountValueGroupBox = true;
            }
        }

        /// <summary>
        /// Processes the order.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <example>
        /// Look on example for <c>SalesDiscountProvider</c> class
        /// </example>
        public override void ProcessOrder(Order order)
        {
            var orderLines = GetDiscountOrderLinesFromCache(order);

            if (orderLines != null)
            {
                foreach (var orderLine in orderLines)
                {
                    if (!order.OrderLines.Contains(orderLine))
                    {
                        order.OrderLines.Add(orderLine.CloneOrderLine());
                    }
                }
            }
        }

        /// <summary>
        /// Adds to cache.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="discountOrderLines">The discount order lines.</param>
        internal static void AddDiscountOrderLinesToCache(Order order, OrderLineCollection discountOrderLines)
        {
            var cacheKey = GetDiscountOrderLinesCacheKey(order);
            Context.Current.Session[cacheKey] = discountOrderLines;
        }

        /// <summary>
        /// Clears the cache.
        /// </summary>
        /// <param name="order">The order.</param>
        internal static void ClearCache(Order order)
        {
            RemoveDiscountOrderLinesFromCache(order);
        }

        /// <summary>
        /// Gets the first live sales discount.
        /// </summary>
        /// <returns>SalesDiscount.</returns>
        internal static SalesDiscount GetSalesDiscount()
        {
            return SalesDiscount.GetSalesDiscounts(Common.Context.Language).FirstOrDefault(sd => !string.IsNullOrEmpty(sd.DiscountType) && string.Compare(sd.DiscountType, typeof(LiveSalesDiscountProvider).FullName, StringComparison.OrdinalIgnoreCase) == 0);
        }

        /// <summary>
        /// Removes discount order lines from cache.
        /// </summary>
        /// <param name="order">The order.</param>
        internal static void RemoveDiscountOrderLinesFromCache(Order order)
        {
            var cacheKey = GetDiscountOrderLinesCacheKey(order);
            Context.Current.Session.Remove(cacheKey);
        }

        /// <summary>
        /// Gets the discount order lines cache key
        /// </summary>
        /// <param name="order">Order used for generating the cache key</param>
        /// <returns>Cache key</returns>
        private static string GetDiscountOrderLinesCacheKey(Order order)
        {
            return "LiveIntegrationDiscounts" + order.Id;
        }

        /// <summary>
        /// Gets discount order lines from cache
        /// </summary>
        /// <param name="order">Order to get discount order lines for</param>
        /// <returns>A collection of discount order lines</returns>
        private static OrderLineCollection GetDiscountOrderLinesFromCache(Order order)
        {
            var cacheKey = GetDiscountOrderLinesCacheKey(order);

            var orderLines = Context.Current.Session[cacheKey] as OrderLineCollection;

            return orderLines;
        }
    }
}