﻿using Dynamicweb.Ecommerce.Cart;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Extensibility.AddIns;
using System;
using System.Linq;
using System.Xml;

namespace Dynamicweb.Ecommerce.LiveIntegration
{
    /// <summary>
    /// Shipping Fee Provider to handle shipping fees coming from the ERP.
    /// </summary>
    /// <seealso cref="ShippingProvider" />
    [AddInName("Live integration shipping fee provider"), AddInDescription("Applies the shipping fee via live integration")]
    public class LiveShippingFeeProvider : ShippingProvider
    {
        internal static bool IsShippingHandledByDynamicweb
        {
            get
            {
                return GetShipping() == null;
            }
        }

        /// <summary>
        /// Calculate shipping fee for the specified order.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <returns>Returns shipping fee for the specified order</returns>
        public override PriceRaw CalculateShippingFee(Order order)
        {
            PriceRaw rate = null;

            if (Context.Current.Session["LiveShippingFeeProvider" + order.Id] != null)
            {
                double shippingFee = (double)Context.Current.Session["LiveShippingFeeProvider" + order.Id];
                rate = new PriceRaw(shippingFee, order.Currency);
            }

            return rate;
        }

        /// <summary>
        /// Clears the cache.
        /// </summary>
        /// <param name="order">The order.</param>
        internal static void ClearCache(Order order)
        {
            Context.Current.Session.Remove("LiveShippingFeeProvider" + order.Id);
        }

        /// <summary>
        /// Processes the shipping.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="orderNode">The order node.</param>
        internal static void ProcessShipping(Order order, XmlNode orderNode)
        {
            if (!IsShippingHandledByDynamicweb)
            {
                ProcessLiveIntegrationShipping(order, orderNode);
            }
        }

        /// <summary>
        /// Adds to cache.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="shippingFee">The shipping fee.</param>
        private static void AddToCache(Order order, double shippingFee)
        {
            Context.Current.Session["LiveShippingFeeProvider" + order.Id] = shippingFee;
        }

        /// <summary>
        /// Gets the shipping.
        /// </summary>
        /// <returns>Shipping.</returns>
        private static Shipping GetShipping()
        {
            return Shipping.GetAllShippingmethods().FirstOrDefault(s => !string.IsNullOrEmpty(s.ServiceSystemName) && string.Compare(s.ServiceSystemName, typeof(LiveShippingFeeProvider).FullName, StringComparison.OrdinalIgnoreCase) == 0);
        }

        private static void ProcessLiveIntegrationShipping(Order order, XmlNode orderNode)
        {
            string shippingFee = orderNode.SelectSingleNode("column [@columnName='OrderShippingFee']")?.InnerText;
            if (!string.IsNullOrEmpty(shippingFee))
            {
                Shipping liveIntegrationShipping = GetShipping();
                order.ShippingMethodId = liveIntegrationShipping.Id;
                string shippingName = orderNode.SelectSingleNode("column [@columnName='OrderShippingMethodName']")?.InnerText;

                if (string.IsNullOrEmpty(shippingName))
                {
                    shippingName = orderNode.SelectSingleNode("column [@columnName='OrderShippingMethodId']")?.InnerText;
                }

                order.ShippingMethod = !string.IsNullOrEmpty(shippingName) ? shippingName : liveIntegrationShipping.Name;

                double fee = Helpers.ToDouble(shippingFee);
                AddToCache(order, fee);

                if (!order.IsCart)
                {
                    order.ShippingFee.PriceWithVAT = fee;
                }
            }
        }
    }
}