﻿using Dynamicweb.Core;
using Dynamicweb.Data;
using Dynamicweb.Ecommerce.Integration;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.LiveIntegration.Extensions;
using Dynamicweb.Ecommerce.LiveIntegration.Logging;
using Dynamicweb.Ecommerce.LiveIntegration.XmlGenerators;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Ecommerce.Prices;
using Dynamicweb.Ecommerce.Products;
using Dynamicweb.Environment;
using Dynamicweb.Extensibility.Notifications;
using Dynamicweb.Security.UserManagement;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml;

namespace Dynamicweb.Ecommerce.LiveIntegration
{
    /// <summary>
    /// Handler class to handle all interaction with the ERP.
    /// </summary>
    internal static class OrderHandler
    {
        private static readonly string OrderErpCallFailed = "OrderHandler.ErpCallFailed";
        private static readonly string OrderErpCallCancelled = "OrderHandler.ErpCallCancelled";
        private static readonly string OrderErpCallSucceed = "OrderHandler.ErpCallSucceed";

        /// <summary>
        /// The order XML log folder
        /// </summary>
        private static readonly string OrderXmlLogFolder = "/Files/System/Log/LiveIntegration/OrderXml";

        /// <summary>
        /// Gets or sets a value that determines if custom order fields are appended to the outgoing XML request.
        /// </summary>
        /// <value>Returns <c>true</c> if order fields should be added to the request; otherwise, <c>false</c>.</value>
        private static bool AddOrderFieldsToRequest => Settings.Instance.AddOrderFieldsToRequest;

        /// <summary>
        /// Gets or sets a value that determines if custom order line fields are appended to the outgoing XML request.
        /// </summary>
        /// <value>Returns <c>true</c> if order line fields should be added to the request; otherwise, <c>false</c>.</value>
        private static bool AddOrderLineFieldsToRequest => Settings.Instance.AddOrderLineFieldsToRequest;

        /// <summary>
        /// Gets or sets the cache level for order information.
        /// </summary>
        /// <value>The order cache level.</value>
        private static CacheLevel OrderCacheLevel
        {
            get
            {
                string cacheLevelString = Settings.Instance.OrderCacheLevel;
                return Helpers.GetEnumValueFromString(cacheLevelString, CacheLevel.Page);
            }
        }

        /// <summary>
        /// Gets a value indicating whether to save a copy of the XML in the logs folder when a completed order is sent to the ERP.
        /// </summary>
        /// <value><c>true</c> if a copy should be saved; otherwise, <c>false</c>.</value>
        private static bool SaveOrderXml => Settings.Instance.SaveCopyOfOrderXml;

        /// <summary>
        /// Updates an order in the ERP.
        /// </summary>
        /// <param name="order">The order that must be synced with the ERP.</param>
        /// <param name="liveIntegrationSubmitType">Determines the origin of this submit such.</param>
        /// <param name="successOrderStateId">The order state that is applied to the order when it integrates successfully.</param>
        /// <param name="failedOrderStateId">The order state that is applied to the order when an error occurred during the integration.</param>
        /// <returns>Returns null if no communication has made, or bool if order has been updated successfully or not.</returns>
        public static bool? UpdateOrder(Order order, SubmitType liveIntegrationSubmitType, string successOrderStateId = null, string failedOrderStateId = null)
        {
            if (!IsOrderUpdateAllowed(order, liveIntegrationSubmitType))
            {
                return null;
            }

            var orderId = order.Id ?? "ID is null";
            Logger.Instance.Log(ErrorLevel.DebugInfo, $"Updating order with ID: {orderId}. Complete: {order.Complete}. Order submitted from the backend: {ExecutingContext.IsBackEnd()}");

            // use current user if is not backend running or if the cart is Anonymous
            var user = User.GetUserByID(order.CustomerAccessUserId) ?? (!ExecutingContext.IsBackEnd() ? User.GetCurrentExtranetUser() : null);

            /* create order: if it is false, you will get a calculate order from the ERP with the total prices */
            /* if it is true, then a new order will be created in the ERP */
            bool createOrder = order.Complete;

            /* Create order if the request is from Backend */
            if (ExecutingContext.IsBackEnd() && !createOrder)
            {
                createOrder = true;
            }            

            if (!Settings.Instance.EnableCartCommunicationForAnonymousUsers && user == null)
            {
                Logger.Instance.Log(ErrorLevel.DebugInfo, $"No user is currently logged in. Anonymous user cart is not allowed. Order = {orderId}");
                return null;
            }

            // default states
            if (successOrderStateId == null)
            {
                successOrderStateId = Settings.Instance.OrderStateAfterExportSucceeded;
            }

            if (failedOrderStateId == null)
            {
                failedOrderStateId = Settings.Instance.OrderStateAfterExportFailed;
            }

            var requestXml = new OrderXmlGenerator().GenerateOrderXml(order, new OrderXmlGeneratorSettings { AddOrderLineFieldsToRequest = AddOrderLineFieldsToRequest, AddOrderFieldsToRequest = AddOrderFieldsToRequest, CreateOrder = createOrder, LiveIntegrationSubmitType = liveIntegrationSubmitType, ReferenceName = "OrdersPut" });

            if (createOrder && SaveOrderXml && liveIntegrationSubmitType == SubmitType.LiveOrderOrCart)
            {
                SaveCopyOfXml(order.Id, requestXml);
            }

            // calculate current hash
            string currentHash = CalculateHash(requestXml);

            // get last hash
            string lastHash = GetLastOrderHash();

            if (!string.IsNullOrEmpty(lastHash) && lastHash == currentHash)
            {
                // no changes to order
                return null;
            }

            // save this hash for next calls
            SaveOrderHash(currentHash);

            // clear all session var to avoid miss calculations
            LiveSalesDiscountProvider.ClearCache(order);
            LiveShippingFeeProvider.ClearCache(order);

            XmlDocument response = GetResponse(requestXml, order, createOrder);
            if (response != null && !string.IsNullOrWhiteSpace(response.InnerXml))
            {
                bool processResponseResult = ProcessResponse(response, order, createOrder, successOrderStateId, failedOrderStateId);                
                return processResponseResult;
            }
            else
            {
                // error occurred
                CheckIfOrderShouldRevertToCart(order, createOrder);

                if (createOrder)
                {
                    OrderDebuggingInfo.Save(order, DateTime.Now, $"ERP communication failed with null response returned.", OrderErpCallFailed, DebuggingInfoType.Undefined);
                }

                return false;
            }
        }

        /// <summary>
        /// Builds the XML copy path.
        /// </summary>
        /// <param name="orderId">The order identifier.</param>
        /// <param name="folder">The folder.</param>
        /// <returns>System.String.</returns>
        internal static string BuildXmlCopyPath(string orderId, string folder)
        {
            return Path.Combine(folder, $"{orderId}.xml");
        }

        /// <summary>
        /// Gets the log folder for XML copies.
        /// </summary>
        /// <param name="date">The date.</param>
        /// <returns>System.String.</returns>
        internal static string GetLogFolderForXmlCopies(DateTime? date = null)
        {
            if (!date.HasValue)
            {
                date = DateTime.Now;
            }

            try
            {
                var logFolder = $"{OrderXmlLogFolder}/{date.Value.Year}/{date.Value:MM}";
                var logFolderPhysical = Context.Current.Server.MapPath(logFolder);

                if (!Directory.Exists(logFolderPhysical))
                {
                    Directory.CreateDirectory(logFolderPhysical);
                }

                return logFolderPhysical;
            }
            catch (Exception e)
            {
                Logger.Instance.Log(ErrorLevel.Error, "Error creating log folder for order XML files: " + e.Message);
                return string.Empty;
            }
        }

        /// <summary>
        /// Assigns the integration order identifier.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="orderNode">The order node.</param>
        private static void AssignIntegrationOrderId(Order order, XmlNode orderNode)
        {
            // search for IntegrationOrderID field in response otherwise use the OrderIntegrationOrderID
            XmlNode integrationIdNode = orderNode.SelectSingleNode("column [@columnName='OrderIntegrationOrderId']");

            // otherwise use the traditional OrderID
            if (string.IsNullOrEmpty(integrationIdNode?.InnerText))
            {
                integrationIdNode = orderNode.SelectSingleNode("column [@columnName='OrderId']");
            }

            if (!string.IsNullOrWhiteSpace(integrationIdNode?.InnerText))
            {
                order.IntegrationOrderId = integrationIdNode.InnerText;
                order.IsExported = true;
            }
        }

        /// <summary>
        /// Checks if the order should be reverted to a cart.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="createOrder">if set to <c>true</c> [create order].</param>
        private static void CheckIfOrderShouldRevertToCart(Order order, bool createOrder)
        {
            if (createOrder && !Settings.Instance.QueueOrdersToExport)
            {
                order.DowngradeToCart();
                Common.Context.SetCart(order);
                order.CartV2StepIndex = --order.CartV2StepIndex;
                order.Complete = false;
            }
        }

        /// <summary>
        /// Creates the order line.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="productId">The product identifier.</param>
        /// <returns>OrderLine.</returns>
        private static OrderLine CreateOrderLine(Order order, string productId)
        {
            if (order == null)
            {
                throw new ArgumentNullException(nameof(order));
            }

            if (string.IsNullOrEmpty(productId))
            {
                throw new ArgumentNullException(nameof(productId));
            }

            var scalar = Converter.ToString(Database.ExecuteScalar($"SELECT TOP 1 ProductID FROM EcomProducts WHERE ProductNumber = '{productId}'"));

            if (string.IsNullOrWhiteSpace(scalar))
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Can not CreateOrderLine: No product found with ProductID = '{productId}'");
                return null;
            }

            var product = Product.GetProductById(scalar);
            if (product == null || product.Id != productId)
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Cannot CreateOrderLine: No product found or ProductNumber is different from the response ProductID = '{productId}' Order = {order.Id}");
                return null;
            }

            OrderLine orderLine = new OrderLine();
            orderLine.SetProductInformation(product);
            orderLine.Type = ((int)OrderLineType.Product).ToString();
            orderLine.Order = order;

            order.OrderLines.Add(orderLine);

            return orderLine;
        }

        /// <summary>
        /// Gets the response.
        /// </summary>
        /// <param name="requestXml">The request XML.</param>
        /// <param name="order">The order.</param>
        /// <param name="createOrder">if set to <c>true</c> [create order].</param>
        /// <returns>XmlDocument.</returns>
        private static XmlDocument GetResponse(string requestXml, Order order, bool createOrder)
        {
            XmlDocument response = null;

            string orderIdentifier = Helpers.OrderIdentifier(order);

            Dictionary<string, XmlDocument> responsesCache = ErpResponseCache.GetWebOrdersConnectorResponses(OrderCacheLevel);

            if (!createOrder && responsesCache.ContainsKey(orderIdentifier))
            {
                response = responsesCache[orderIdentifier];
            }
            else
            {
                Notifications.Order.OnBeforeSendingOrderToErpArgs onBeforeSendingOrderToErpArgs = new Notifications.Order.OnBeforeSendingOrderToErpArgs(order, createOrder);
                NotificationManager.Notify(Notifications.Order.OnBeforeSendingOrderToErp, onBeforeSendingOrderToErpArgs);

                if (!onBeforeSendingOrderToErpArgs.Cancel)
                {
                    Exception error = null;
                    response = Connector.CalculateOrder(requestXml, order.Id, createOrder, out error);

                    if (createOrder && error != null)
                    {
                        OrderDebuggingInfo.Save(order, DateTime.Now, $"ERP communication failed with error: {error}", OrderErpCallFailed, DebuggingInfoType.Undefined);
                    }

                    NotificationManager.Notify(Notifications.Order.OnAfterSendingOrderToErp, new Notifications.Order.OnAfterSendingOrderToErpArgs(order, createOrder, response, error));

                    if (responsesCache.ContainsKey(orderIdentifier))
                    {
                        responsesCache.Remove(orderIdentifier);
                    }

                    if (response != null && !string.IsNullOrWhiteSpace(response.InnerXml))
                    {
                        responsesCache.Add(orderIdentifier, response);
                    }
                }
                else
                {
                    OrderDebuggingInfo.Save(order, DateTime.Now, "Order not sent to ERP because a subscriber cancelled sending it", OrderErpCallCancelled, DebuggingInfoType.Undefined);
                }
            }

            return response;
        }

        /// <summary>
        /// Handles the integration failure.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="failedState">State of the failed.</param>
        /// <param name="orderId">The order identifier.</param>
        /// <param name="discountOrderLines">The discount order lines.</param>
        private static void HandleIntegrationFailure(Order order, string failedState, string orderId, OrderLineCollection discountOrderLines)
        {
            if (Global.EnableCartCommunication(order.Complete))
            {
                RemoveDiscounts(order);
                order.OrderLines.Add(discountOrderLines);
            }

            Logger.Instance.Log(ErrorLevel.Error, $"Order with ID '{orderId}' was not created in the ERP system.");
            Context.Current.Session["LiveIntegration.OrderExportFailed"] = true;
            Context.Current.Session["LiveIntegration.FailedOrderId"] = order.Id;

            if (!Settings.Instance.QueueOrdersToExport)
            {
                order.DowngradeToCart();
                Common.Context.SetCart(order);
                order.CartV2StepIndex = --order.CartV2StepIndex;
                order.Complete = false;
            }

            if (!string.IsNullOrWhiteSpace(failedState))
            {
                order.StateId = failedState;
            }

            order.Save();
        }

        /// <summary>
        /// Handles the integration success.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="successState">State of the success.</param>
        private static void HandleIntegrationSuccess(Order order, string successState)
        {
            if (!string.IsNullOrWhiteSpace(successState))
            {
                order.StateId = successState;
            }

            order.Save();
            Context.Current.Session["LiveIntegration.OrderExportFailed"] = null;
            Context.Current.Session["LiveIntegration.FailedOrderId"] = null;
        }

        private static bool IsOrderUpdateAllowed(Order order, SubmitType liveIntegrationSubmitType)
        {
            if (order == null
                || !order.OrderLines.Any()
                || (order.IsLedgerEntry && Settings.Instance.SkipLedgerOrder)
                || (liveIntegrationSubmitType == SubmitType.LiveOrderOrCart && !string.IsNullOrEmpty(order.IntegrationOrderId)))
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// Sets the price of the order with the values from the ERP.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="orderNode">The order node.</param>
        private static void OrderPriceCalculation(Order order, XmlNode orderNode)
        {
            order.Price.Currency = order.Currency;
            var orderPriceNode = orderNode.SelectSingleNode("column [@columnName='OrderPrice']");

            if (orderPriceNode == null)
            {
                orderPriceNode = orderNode.SelectSingleNode("column [@columnName='OrderPriceWithVat']");
            }

            if (orderPriceNode != null)
            {
                order.Price.PriceWithVAT = Helpers.ToDouble(orderPriceNode.InnerText);
            }

            var orderPriceWithoutVatNode = orderNode.SelectSingleNode("column [@columnName='OrderPriceWithoutVat']");

            if (orderPriceWithoutVatNode != null)
            {
                order.Price.PriceWithoutVAT = Helpers.ToDouble(orderPriceWithoutVatNode.InnerText);
            }

            if (order.Price.PriceWithVAT > 0 && order.Price.PriceWithoutVAT > 0)
            {
                order.Price.VAT = order.Price.PriceWithVAT - order.Price.PriceWithoutVAT;
            }

            var orderPriceBeforeFeesWithVat = orderNode.SelectSingleNode("column [@columnName='OrderPriceBeforeFeesWithVat']");

            if (orderPriceBeforeFeesWithVat != null)
            {
                order.PriceBeforeFees.PriceWithVAT = Helpers.ToDouble(orderPriceBeforeFeesWithVat.InnerText);
            }
            else
            {
                order.PriceBeforeFees.PriceWithVAT = order.Price.PriceWithVAT;
            }

            var orderPriceBeforeFeesWithoutVat = orderNode.SelectSingleNode("column [@columnName='OrderPriceBeforeFeesWithoutVat']");

            if (orderPriceBeforeFeesWithoutVat != null)
            {
                order.PriceBeforeFees.PriceWithoutVAT = Helpers.ToDouble(orderPriceBeforeFeesWithoutVat.InnerText);
            }
            else
            {
                order.PriceBeforeFees.PriceWithoutVAT = order.Price.PriceWithVAT;
            }

            if (order.PriceBeforeFees.PriceWithVAT > 0 && order.PriceBeforeFees.PriceWithoutVAT > 0)
            {
                order.PriceBeforeFees.VAT = order.PriceBeforeFees.PriceWithVAT - order.PriceBeforeFees.PriceWithoutVAT;
            }
        }

        /// <summary>
        /// Processes the discount order line.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="discountOrderLines">The discount order lines.</param>
        /// <param name="orderLineNode">The order line node.</param>
        /// <param name="orderLineType">Type of the order line.</param>
        private static void ProcessDiscountOrderLine(Order order, OrderLineCollection discountOrderLines, XmlNode orderLineNode, string orderLineType)
        {
            string orderLineId = orderLineNode.SelectSingleNode("column [@columnName='OrderLineId']")?.InnerText;

            try
            {
                var orderLine = new OrderLine();
                orderLine.Order = order;
                orderLine.ProductName = Settings.Instance.TextForDiscounts;
                orderLine.Type = orderLineType;

                if (orderLine.IsProductDiscount())
                {
                    string parentProductId = orderLineNode.SelectSingleNode("column [@columnName='OrderLineProductNumber']")?.InnerText;
                    if (!string.IsNullOrEmpty(parentProductId))
                    {
                        foreach (var productOrderLine in order.OrderLines)
                        {
                            if (productOrderLine.ProductId == parentProductId)
                            {
                                orderLine.ParentLineId = productOrderLine.Id;
                            }
                        }
                    }
                }

                Orders.SalesDiscounts.SalesDiscount liveIntegrationDiscount = LiveSalesDiscountProvider.GetSalesDiscount();
                if (liveIntegrationDiscount != null)
                {
                    orderLine.DiscountId = liveIntegrationDiscount.Id;
                }

                string discountName = orderLineNode.SelectSingleNode("column [@columnName='OrderLineProductName']")?.InnerText;
                if (!string.IsNullOrEmpty(discountName))
                {
                    orderLine.ProductName = discountName;
                }

                double? value = ReadDouble(orderLineNode, "column [@columnName='OrderLineQuantity']");
                if (value.HasValue)
                {
                    orderLine.Quantity = value.Value;
                }
                else
                {
                    orderLine.Quantity = 1;
                }

                SetPrice(
                    orderLine.UnitPrice,
                    ReadDouble(orderLineNode, "column [@columnName='OrderLineUnitPriceWithVat']"),
                    ReadDouble(orderLineNode, "column [@columnName='OrderLineUnitPriceWithoutVat']"),
                    ReadDouble(orderLineNode, "column [@columnName='OrderLineUnitPriceVat']"));

                SetPrice(
                    orderLine.Price,
                    ReadDouble(orderLineNode, "column [@columnName='OrderLinePriceWithVat']"),
                    ReadDouble(orderLineNode, "column [@columnName='OrderLinePriceWithoutVat']"),
                    ReadDouble(orderLineNode, "column [@columnName='OrderLinePriceVat']"));

                orderLine.UnitPrice.PriceWithoutVAT = orderLine.UnitPrice.PriceWithoutVAT > 0 ? -orderLine.UnitPrice.PriceWithoutVAT : orderLine.UnitPrice.PriceWithoutVAT;
                orderLine.UnitPrice.PriceWithVAT = orderLine.UnitPrice.PriceWithVAT > 0 ? -orderLine.UnitPrice.PriceWithVAT : orderLine.UnitPrice.PriceWithVAT;
                orderLine.Price.PriceWithVAT = orderLine.Price.PriceWithVAT > 0 ? -orderLine.Price.PriceWithVAT : orderLine.Price.PriceWithVAT;
                orderLine.Price.PriceWithoutVAT = orderLine.Price.PriceWithoutVAT > 0 ? -orderLine.Price.PriceWithoutVAT : orderLine.Price.PriceWithoutVAT;

                discountOrderLines.Add(orderLine);
            }
            catch (Exception ex)
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Error processing order line. Error: '{ex.Message}' OrderLineId = {orderLineId}.");
                throw;
            }
        }

        /// <summary>
        /// Processes the order lines.
        /// </summary>
        /// <param name="response">The response.</param>
        /// <param name="order">The order.</param>
        /// <param name="discountOrderLines">The discount order lines.</param>
        private static void ProcessOrderLines(XmlDocument response, Order order, OrderLineCollection discountOrderLines)
        {
            XmlNodeList orderLinesNodes = response.SelectNodes("//item [@table='EcomOrderLines']");

            // Process OrderLines
            if (orderLinesNodes != null && orderLinesNodes.Count > 0)
            {
                List<string> orderLineIds = new List<string>();
                List<OrderLine> orderLines = order.OrderLines.ToList();
                OrderLineFieldCollection allOrderLineFields = null;
                if (AddOrderLineFieldsToRequest)
                {
                    allOrderLineFields = OrderLineField.GetOrderLineFields();
                }

                foreach (XmlNode orderLineNode in orderLinesNodes)
                {
                    XmlNode orderLineTypeNode = orderLineNode.SelectSingleNode("column [@columnName='OrderLineType']");

                    if (orderLineTypeNode == null)
                    {
                        orderLineTypeNode = orderLineNode.SelectSingleNode("column [@columnName='OrderLineTypeId']");
                    }

                    string orderLineType = orderLineTypeNode?.InnerText;
                    if (orderLineType == "0" || string.IsNullOrWhiteSpace(orderLineType))
                    {
                        ProcessProductOrderLine(order, orderLineIds, orderLines, allOrderLineFields, orderLineNode);
                    }

                    // 1=order discount, 3=Product Discount
                    if (orderLineType == "1" || orderLineType == "3")
                    {
                        ProcessDiscountOrderLine(order, discountOrderLines, orderLineNode, orderLineType);
                    }
                }

                // Remove deleted OrderLines
                for (int i = order.OrderLines.Count - 1; i >= 0; i--)
                {
                    var orderLine = order.OrderLines[i];
                    if (string.IsNullOrWhiteSpace(orderLine.Id) ||
                        orderLineIds.Contains(orderLine.Id))
                    {
                        continue;
                    }

                    order.OrderLines.Remove(orderLine);
                    orderLine.Delete();
                }
            }
        }

        /// <summary>
        /// Processes the product order line.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="orderLineIds">The order line IDs.</param>
        /// <param name="orderLines">The order lines.</param>
        /// <param name="allOrderLineFields">All order line fields.</param>
        /// <param name="orderLineNode">The order line node.</param>
        private static void ProcessProductOrderLine(Order order, List<string> orderLineIds, List<OrderLine> orderLines, OrderLineFieldCollection allOrderLineFields, XmlNode orderLineNode)
        {
            string productNumber = orderLineNode.SelectSingleNode("column [@columnName='OrderLineProductNumber']")?.InnerText;

            try
            {
                if (!string.IsNullOrWhiteSpace(productNumber))
                {
                    OrderLine orderLine = orderLines.FirstOrDefault(ol => ol.ProductNumber == productNumber);
                    if (orderLine != null)
                    {
                        // Remove found line for getting next line with same ProductNumber
                        orderLines.Remove(orderLine);
                    }
                    else
                    {
                        // Create an OrderLine if it doesn't exist
                        orderLine = CreateOrderLine(order, productNumber);

                        if (orderLine == null)
                        {
                            return;
                        }
                    }

                    if (!string.IsNullOrWhiteSpace(orderLine.Id))
                    {
                        orderLineIds.Add(orderLine.Id);
                    }

                    // Set standard values on OrderLines
                    orderLine.AllowOverridePrices = true;

                    var doubleValue = ReadDouble(orderLineNode, "column [@columnName='OrderLineQuantity']");
                    if (doubleValue.HasValue)
                    {
                        orderLine.Quantity = doubleValue.Value;
                    }
                    var unitId = orderLineNode.SelectSingleNode("column [@columnName='OrderLineUnitId']")?.InnerText;
                    if (!string.IsNullOrWhiteSpace(unitId))
                    {
                        orderLine.UnitId = unitId;
                    }

                    // order line unit price
                    PriceInfo unitPrice = new PriceInfo(order.Currency);
                    SetPrice(
                        unitPrice,
                        ReadDouble(orderLineNode, "column [@columnName='OrderLineUnitPriceWithVat']"),
                        ReadDouble(orderLineNode, "column [@columnName='OrderLineUnitPriceWithoutVat']"),
                        ReadDouble(orderLineNode, "column [@columnName='OrderLineUnitPriceVat']"));

                    orderLine.SetUnitPrice(unitPrice, false);
                    orderLine.Type = Convert.ToString(Convert.ToInt32(OrderLineType.Fixed));

                    // order line price
                    SetPrice(
                        orderLine.Price,
                        ReadDouble(orderLineNode, "column [@columnName='OrderLinePriceWithVat']"),
                        ReadDouble(orderLineNode, "column [@columnName='OrderLinePriceWithoutVat']"),
                        ReadDouble(orderLineNode, "column [@columnName='OrderLinePriceVat']"));

                    // Set OrderLineCustomFields values
                    if (AddOrderLineFieldsToRequest && allOrderLineFields != null && allOrderLineFields.Count > 0)
                    {
                        XmlNode orderLineFieldNode = null;
                        foreach (OrderLineField olf in allOrderLineFields)
                        {
                            orderLineFieldNode = orderLineNode.SelectSingleNode(
                                $"column [@columnName='{olf.SystemName}']");
                            if (orderLineFieldNode != null)
                            {
                                if (orderLine.OrderLineFieldValues == null)
                                {
                                    orderLine.OrderLineFieldValues = new OrderLineFieldValueCollection();
                                }

                                OrderLineFieldValue olfv = orderLine.OrderLineFieldValues.FirstOrDefault(fv => fv != null && string.Compare(olf.SystemName, fv.OrderLineFieldSystemName, StringComparison.OrdinalIgnoreCase) == 0);
                                if (olfv != null)
                                {
                                    olfv.Value = orderLineFieldNode.InnerText;
                                }
                                else
                                {
                                    olfv = new OrderLineFieldValue(olf.SystemName, orderLineFieldNode.InnerText);
                                    orderLine.OrderLineFieldValues.Add(olfv);
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Error processing order line. Error: '{ex.Message}' productNumber = {productNumber}.");
                throw;
            }
        }

        /// <summary>
        /// Processes the response.
        /// </summary>
        /// <param name="response">The response.</param>
        /// <param name="order">The order.</param>
        /// <param name="createOrder">if set to <c>true</c> [create order].</param>
        /// <param name="successState">State of the success.</param>
        /// <param name="failedState">State of the failed.</param>
        /// <returns><c>true</c> if response was processed successfully, <c>false</c> otherwise.</returns>
        private static bool ProcessResponse(XmlDocument response, Order order, bool createOrder, string successState, string failedState)
        {
            var orderId = order == null ? "is null" : order.Id ?? "ID is null";
            if (response == null || order == null)
            {
                if (createOrder)
                {
                    // if must create order and no response or invalid order fail to sync
                    Logger.Instance.Log(ErrorLevel.Error, $"Response CreateOrder is null. Order = {orderId}");
                    return false;
                }

                // nothing to do so work done
                return true;
            }

            try
            {
                XmlNode orderNode = response.SelectSingleNode("//item [@table='EcomOrders']");
                if (Global.EnableCartCommunication(order.Complete))
                {
                    // Set Order prices
                    order.AllowOverridePrices = true;
                    try
                    {
                        SetPrices(order, orderNode);

                        SetCustomerNumber(order, orderNode);
                    }
                    catch (Exception ex)
                    {
                        Logger.Instance.Log(ErrorLevel.Error, $"Exception setting prices: {ex.Message} Order = {orderId}");
                    }
                }

                SetCustomOrderFields(order, orderNode);

                var discountOrderLines = new OrderLineCollection();

                if (Global.EnableCartCommunication(order.Complete))
                {
                    ProcessOrderLines(response, order, discountOrderLines);
                    order.OrderLines.Add(discountOrderLines);
                    LiveShippingFeeProvider.ProcessShipping(order, orderNode);
                }

                if (createOrder)
                {
                    AssignIntegrationOrderId(order, orderNode);
                    bool orderCreatedSuccessfully;
                    bool.TryParse(orderNode?.SelectSingleNode("column [@columnName='OrderCreated']")?.InnerText, out orderCreatedSuccessfully);

                    if (!orderCreatedSuccessfully)
                    {
                        HandleIntegrationFailure(order, failedState, orderId, discountOrderLines);
                    }
                    else
                    {
                        HandleIntegrationSuccess(order, successState);
                    }
                }

                if (Global.EnableCartCommunication(order.Complete))
                {
                    // add discount order lines to cache, to be added by the discountProvider
                    LiveSalesDiscountProvider.AddDiscountOrderLinesToCache(order, discountOrderLines);
                }
            }
            catch (Exception ex)
            {
                Logger.Instance.Log(ErrorLevel.Error, $"Error processing response. Error: '{ex.Message}' Order = {orderId}.");

                if (createOrder)
                {
                    OrderDebuggingInfo.Save(order, DateTime.Now, $"ERP communication failed with error: {ex}", OrderErpCallFailed, DebuggingInfoType.Undefined);
                }

                return false;
            }

            if (createOrder)
            {
                OrderDebuggingInfo.Save(order, DateTime.Now, $"Order saved in ERP successfully.", OrderErpCallSucceed, DebuggingInfoType.Undefined);
            }

            return true;
        }

        /// <summary>
        /// Reads a double from the XML node.
        /// </summary>
        /// <param name="elementNode">The element node.</param>
        /// <param name="field">The field.</param>
        /// <returns>System.Nullable&lt;System.Double&gt;.</returns>
        private static double? ReadDouble(XmlNode elementNode, string field)
        {
            XmlNode xValue = elementNode.SelectSingleNode(field);

            if (xValue == null)
            {
                return null;
            }

            return Helpers.ToDouble(xValue.InnerText);
        }

        /// <summary>
        /// Removes the discounts.
        /// </summary>
        /// <param name="order">The order.</param>
        private static void RemoveDiscounts(Order order)
        {
            foreach (var orderLine in order.OrderLines)
            {
                if (orderLine.IsDiscount())
                {
                    order.OrderLines.Remove(orderLine);
                    orderLine.Delete();
                }
            }
        }

        /// <summary>
        /// Saves a copy of the order XML to a custom log folder under the main LiveIntegration log folder.
        /// </summary>
        /// <param name="orderId">The ID of the order being saved; used in the file name.</param>
        /// <param name="requestXml">The order XML to save.</param>
        private static void SaveCopyOfXml(string orderId, string requestXml)
        {
            try
            {
                var folder = GetLogFolderForXmlCopies();
                string file = BuildXmlCopyPath(orderId, folder);
                var doc = new XmlDocument();
                doc.LoadXml(requestXml);
                File.WriteAllText(file, doc.Beautify());
            }
            catch (Exception e)
            {
                Logger.Instance.Log(ErrorLevel.Error, "Error writing copy of XML to log folder: " + e.Message);
            }
        }

        /// <summary>
        /// Sets the customer number.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="orderNode">The order node.</param>
        private static void SetCustomerNumber(Order order, XmlNode orderNode)
        {
            var orderCustomerNumber = orderNode.SelectSingleNode("column [@columnName='OrderCustomerNumber']");
            if (orderCustomerNumber != null)
            {
                var newCustomerNumber = orderCustomerNumber.InnerText;

                if (string.Compare(order.CustomerNumber, newCustomerNumber, StringComparison.InvariantCultureIgnoreCase) != 0)
                {
                    order.CustomerNumber = newCustomerNumber;
                }
            }
        }

        /// <summary>
        /// Sets the custom order fields.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="orderNode">The order node.</param>
        private static void SetCustomOrderFields(Order order, XmlNode orderNode)
        {
            // Set OrderCustomFields values
            if (!AddOrderFieldsToRequest || order.OrderFieldValues.Count <= 0)
            {
                return;
            }

            foreach (var orderFieldValue in order.OrderFieldValues)
            {
                var fieldNode = orderNode.SelectSingleNode(
                    $"column [@columnName='{orderFieldValue.OrderField.SystemName}']");
                if (fieldNode != null)
                {
                    orderFieldValue.Value = fieldNode.InnerText;
                }
            }
        }

        private static void SetPrice(PriceInfo price, double? priceWithVat, double? priceWithoutVat, double? priceVat)
        {
            if (priceWithVat.HasValue && priceWithoutVat.HasValue && priceVat.HasValue)
            {
                price.VAT = priceVat.Value;
                price.PriceWithVAT = priceWithVat.Value;
                price.PriceWithoutVAT = priceWithoutVat.Value;
            }
            else
            {
                if (priceVat.HasValue)
                {
                    price.VAT = priceVat.Value;
                }

                if (priceWithVat.HasValue)
                {
                    price.PriceWithVAT = priceWithVat.Value;
                    price.PriceWithoutVAT = price.PriceWithVAT - price.VAT;
                }

                if (priceWithoutVat.HasValue)
                {
                    price.PriceWithoutVAT = priceWithoutVat.Value;
                    price.PriceWithVAT = price.PriceWithoutVAT + price.VAT;
                }
            }
        }

        /// <summary>
        /// Sets the prices.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <param name="orderNode">The order node.</param>
        private static void SetPrices(Order order, XmlNode orderNode)
        {
            OrderPriceCalculation(order, orderNode);
            var orderDiscount = orderNode.SelectSingleNode("column [@columnName='OrderSalesDiscount']");
            if (orderDiscount != null)
            {
                order.SalesDiscount = Helpers.ToDouble(orderDiscount.InnerText);
            }
        }        

        #region Session Hash helper methods

        /// <summary>
        /// Calculates the hash.
        /// </summary>
        /// <param name="orderRequest">The order request.</param>
        /// <returns>System.String.</returns>
        private static string CalculateHash(string orderRequest)
        {
            using (var hashEngine = new System.Security.Cryptography.HMACMD5(new byte[] { 14, 4, 78 }))
            {
                byte[] hash = hashEngine.ComputeHash(Encoding.UTF8.GetBytes(orderRequest));
                string hashString = string.Empty;

                foreach (byte x in hash)
                {
                    hashString += $"{x:x2}";
                }

                return hashString;
            }
        }

        /// <summary>
        /// Gets the last order hash.
        /// </summary>
        /// <returns>System.String.</returns>
        private static string GetLastOrderHash()
        {
            string hashValue = null;
            object orderHash = Context.Current.Session[GetSessionCacheKey()];

            if (orderHash is string)
            {
                hashValue = orderHash as string;
            }

            return hashValue;
        }

        /// <summary>
        /// Gets the session cache key.
        /// </summary>
        /// <returns>System.String.</returns>
        private static string GetSessionCacheKey()
        {
            var user = User.GetCurrentExtranetUser();
            string sessionCache = Constants.CacheConfiguration.OrderCommunicationHash;

            if (user != null)
            {
                sessionCache += user.ID.ToString();
            }
            else
            {
                sessionCache += Settings.Instance.AnonymousUserKey;
            }

            return sessionCache;
        }

        /// <summary>
        /// Saves the order hash.
        /// </summary>
        /// <param name="hashValue">The hash value.</param>
        private static void SaveOrderHash(string hashValue)
        {
            Context.Current.Session[GetSessionCacheKey()] = hashValue;
        }

        #endregion Session Hash helper methods
    }
}