﻿using Dynamicweb.Controls;
using Dynamicweb.Controls.Extensibility;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Ecommerce.Orders;
using Dynamicweb.Extensibility.AddIns;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Web.UI;

namespace Dynamicweb.Ecommerce.LiveIntegration.RibbonBars
{
    /// <summary>
    /// Class SendOrderRibbonBar.
    /// </summary>
    /// <seealso cref="RibbonBarBase" />
    [AddInName("Send order")]
    [AddInDescription("Send order to ERP")]
    [AddInTarget(RibbonBarAddInTarget.eCom.OrderEdit)]
    [AddInTarget(RibbonBarAddInTarget.eCom.OrderList)]
    public class SendOrderRibbonBar : RibbonBarBase
    {
        /// <summary>
        /// Initializes a new instance of an object.
        /// </summary>
        /// <param name="ribbon">An instance of <see cref="T:Dynamicweb.Controls.RibbonBar">RibbonBar</see> object that contains an add-in.</param>
        public SendOrderRibbonBar(RibbonBar ribbon) : base(ribbon)
        {
        }

        /// <summary>
        /// Gets the session key.
        /// </summary>
        /// <value>The session key.</value>
        private string SessionKey
        {
            get
            {
                string key = "LiveIntegration.TransferredOrder";
                Order order = Ribbon.DataContext.DataSource as Order;

                if (order != null)
                {
                    key += order.Id;
                }

                return key;
            }
        }

        /// <summary>
        /// Fires when this ribbon bar add in loads. Adds a transfer button to the Orders section.
        /// </summary>
        public override void Load()
        {
            using (var button = CreateTransferOrderButton(CreateLiveIntegrationRibbon("Orders")))
            {
                if (Settings.Instance.IsLiveIntegrationEnabled)
                {
                    button.EnableServerClick = true;

                    // If Order list is loaded
                    var source = Ribbon.DataContext.DataSource as Order[];
                    if (source != null)
                    {
                        var orders = source;
                        if (orders.Any(o => o != null && string.IsNullOrEmpty(o.IntegrationOrderId)))
                        {
                            button.OnClientClick = "if(!confirm('Transfer selected orders to ERP?')) { return false; };";
                            button.Click += TransferOrdersButton_Click;
                        }
                        else
                        {
                            button.OnClientClick = "alert('All selected orders were already transferred');";
                        }
                    }
                    else
                    {
                        // If Order edit is loaded
                        if (Ribbon.DataContext.DataSource is Order)
                        {
                            // Access the DataContext which is an Order for this AddIn.
                            var order = (Order)Ribbon.DataContext.DataSource;
                            if (string.IsNullOrEmpty(order.IntegrationOrderId))
                            {
                                button.OnClientClick =
                                    $"if(!confirm('Transfer order {order.Id} to ERP?')) {{ return false; }};";
                                button.Click += TransferOrderButton_Click;
                            }
                            else
                            {
                                button.OnClientClick =
                                    $"if(!confirm('Order {order.Id} is already in the ERP, update again?')) {{ return false; }};";
                                button.Click += TransferOrderButton_Click;
                            }
                        }
                    }
                }
                else
                {
                    button.OnClientClick = "alert('Live integration is not enabled');";
                }
            }
        }

        /// <summary>
        /// Sends add-in's output to specified <see cref="T:System.Web.UI.HtmlTextWriter">HtmlTextWriter</see> object.
        /// </summary>
        /// <param name="writer">An instance of <see cref="T:System.Web.UI.HtmlTextWriter">HtmlTextWriter</see> object to send output to.</param>
        public override void Render(HtmlTextWriter writer)
        {
            if (Context.Current.Session[SessionKey] != null && !string.IsNullOrEmpty(Context.Current.Session[SessionKey].ToString()))
            {
                writer.Write("<script>alert('{0}')</script>", System.Web.HttpUtility.HtmlEncode(Context.Current.Session[SessionKey]));
                Context.Current.Session.Remove(SessionKey);
            }

            base.Render(writer);
        }

        /// <summary>
        /// Gets the exported orders message.
        /// </summary>
        /// <param name="rows">The rows.</param>
        /// <param name="exportedOrders">The exported orders.</param>
        /// <param name="alreadyExportedOrders">The already exported orders.</param>
        /// <returns>System.String.</returns>
        private string GetExportedOrdersMessage(ListRowCollection rows, List<string> exportedOrders, List<string> alreadyExportedOrders)
        {
            string output = string.Empty;
            if (alreadyExportedOrders.Count > 0 && alreadyExportedOrders.Count == rows.Count(r => r.Selected))
            {
                output = "All selected orders are already transferred to ERP.";
            }
            else if (exportedOrders.Count > 0 || alreadyExportedOrders.Count > 0)
            {
                if ((exportedOrders.Count + alreadyExportedOrders.Count) == rows.Count(r => r.Selected))
                {
                    output = "All selected orders were successfully transferred to ERP.";
                }
                else
                {
                    if (alreadyExportedOrders.Count > 0)
                    {
                        output += $"Orders with IDs [{string.Join(",", alreadyExportedOrders)}] were already transferred to ERP. ";
                    }

                    if (exportedOrders.Count > 0)
                    {
                        output += $"Orders with IDs [{string.Join(",", exportedOrders)}] were successfully transferred to ERP. ";
                    }

                    output += $"Orders with IDs [{string.Join(",", rows.Where(r => r.Selected && !(exportedOrders.Contains(r.ItemID) || alreadyExportedOrders.Contains(r.ItemID))).Select(r => r.ItemID).Distinct().ToArray())}] were not transferred to ERP. See LiveIntegration log for details";
                }
            }
            else
            {
                output = "None of the selected orders were transferred to ERP. See LiveIntegration log for details.";
            }

            return output;
        }

        /// <summary>
        /// Transfers the order.
        /// </summary>
        /// <param name="order">The order.</param>
        /// <returns><c>true</c> if the order was transferred, <c>false</c> otherwise.</returns>
        private bool TransferOrder(Order order)
        {
            bool? result = OrderHandler.UpdateOrder(order, SubmitType.ManualSubmit);
            return result.HasValue && result.Value;
        }

        /// <summary>
        /// Occurs when the button was clicked from edit order page
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void TransferOrderButton_Click(object sender, EventArgs e)
        {
            string output = string.Empty;

            Order order = Ribbon.DataContext.DataSource as Order;
            if (order != null)
            {
                if (string.IsNullOrEmpty(order.IntegrationOrderId))
                {
                    if (TransferOrder(order))
                    {
                        output = "Order successfully transferred to ERP";
                    }
                    else
                    {
                        output = "Error creating order in ERP. See LiveIntegration log for details";
                    }
                }
                else
                {
                    output = "Order already transferred";
                }
            }

            Context.Current.Session[SessionKey] = output;
        }

        /// <summary>
        /// Occurs when the button was clicked from orders list page
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="EventArgs"/> instance containing the event data.</param>
        private void TransferOrdersButton_Click(object sender, EventArgs e)
        {
            string output;

            var rows = Context.Current.Session["DW.Controls.Tree.List"] as ListRowCollection;
            if (rows != null)
            {
                // Get selected orders from order list                
                if (rows.Count > 0 && rows.Any(r => r.Selected))
                {
                    List<string> exportedOrders = new List<string>();
                    List<string> alreadyExportedOrders = new List<string>();

                    foreach (string id in rows.Where(r => r.Selected).Select(r => r.ItemID).Distinct())
                    {
                        Order o = Order.GetOrderById(id);
                        if (o != null)
                        {
                            if (string.IsNullOrEmpty(o.IntegrationOrderId))
                            {
                                if (TransferOrder(o))
                                {
                                    exportedOrders.Add(o.Id);
                                }
                            }
                            else
                            {
                                alreadyExportedOrders.Add(o.Id);
                            }
                        }
                    }

                    output = GetExportedOrdersMessage(rows, exportedOrders, alreadyExportedOrders);
                }
                else
                {
                    output = "No orders selected for transfer";
                }
            }
            else
            {
                output = "Can not load selected orders from the order list";
            }

            Context.Current.Session[SessionKey] = output;
        }
    }
}